/* compute.c */
/* Copyright (c) 2000-2003 Fabien Tassin <fta@sofaraway.org> */

#include <string.h>
#include <stdio.h>
#include "innpeer-stats.h"

extern char   **f;
extern time_t   max_date, min_date;

void compute(stat_feed_t ***data, stat_feed_t *old, stat_feed_t *new) {
  stat_feed_t   cur, **d, **p;
  time_t           t;
  double	   dt;
  int              i, j;

  cur = *new;
  if (old != NULL) {
    cur.seconds   -= old->seconds;
    cur.offered   -= old->offered;
    cur.accepted  -= old->accepted;
    cur.refused   -= old->refused;
    cur.rejected  -= old->rejected;
    cur.duplicate -= old->duplicate;
    cur.spooled   -= old->spooled;
    cur.missing   -= old->missing;
    cur.accsize   -= old->accsize;
    cur.rejsize   -= old->rejsize;
    /* nothing to do for cur.queue.* */
  }
  t = cur.date - cur.seconds;
  if (t < min_date) /* mainly for innd */
    t = min_date;
  t -= t % INTERVAL;
  for (i = 0, d = *data; *d != NULL && (*d)->date < t; d++, i++);
  if (*d == NULL) {
    int k = 0;
    int date = t;
    if (i > 0 && *data != NULL) {
      k = (*(d - 1))->date;
      k -= k % INTERVAL;
      k = (t - k) / INTERVAL + 1;
      date = (*(d - 1))->date + INTERVAL;
    }
    *data = realloc(*data, (i + 3 + k + (cur.date - t) / INTERVAL) *
		   sizeof(stat_feed_t *));
    for (j = i; j <= k + i + 1 + (cur.date - t) / INTERVAL; j++) {
      (*data)[j] = calloc(1, sizeof(stat_feed_t));
      (*data)[j]->date = date;
      date += INTERVAL;
    }
    (*data)[j] = NULL;
    for (d = *data; *d != NULL && (*d)->date < t; d++);
  }
  else {
    int k = cur.date - cur.date % INTERVAL;
    int date;
    for (i = 0, p = *data; *p != NULL && (*p)->date < k; p++, i++);
    if (*p == NULL) {
      /* i + 3 instead of i + 2 to avoid troubles when cur.seconds modulus
	 INTERVAL equal to zero */
      *data = realloc(*data, (i + 3 + (int) ((cur.date - t) / INTERVAL)) *
		      sizeof(stat_feed_t *));
      date = (*data)[i - 1]->date + INTERVAL;
      for (j = i; j < i + 2 + (cur.date - t) / INTERVAL; j++) {
	(*data)[j] = calloc(1, sizeof(stat_feed_t));
	(*data)[j]->date = date;
	date += INTERVAL;
      }
      (*data)[j] = NULL;
      for (d = *data; *d != NULL && (*d)->date < t; d++);
    }
  }
  while (t <= cur.date) {
    dt = (cur.date > t + INTERVAL) ? t + INTERVAL : cur.date;
    dt -= (t > cur.date - cur.seconds) ? t : cur.date - cur.seconds;
    if (*d == NULL) {
      /* printf("Content-type: text/html\n\n"
	     "<B>This should not happen. FIX ME ! %s:%d </B>\n",
	     __FILE__, __LINE__); */
      return;
    }
    (*d)->sec += dt;
    (*d)->count++;
    if (cur.seconds != 0)
      dt /= cur.seconds;
    else
      dt = 1;
    (*d)->offered   += cur.offered   * dt;
    (*d)->accepted  += cur.accepted  * dt;
    (*d)->refused   += cur.refused   * dt;
    (*d)->rejected  += cur.rejected  * dt;
    (*d)->spooled   += cur.spooled   * dt;
    (*d)->missing   += cur.missing   * dt;
    (*d)->duplicate += cur.duplicate * dt;
    (*d)->accsize   += cur.accsize   * dt;
    (*d)->rejsize   += cur.rejsize   * dt;
    d++;
    t += INTERVAL;
    i++;
  }
}

char *compute_total(stat_peers_t **stat_peers, char way) {
  stat_peers_t     **p;
  stat_feed_t      **q, **r, **data;
  time_t             min, max, t;
  char               buf[BUFSIZE];
  int                i, ok;

  /* get the minimum and maximum dates */
  min = 2147483647UL;
  max = 0;
  ok = 0;
  for (p = stat_peers; *p != NULL; p++) {
    if ((*p)->way != way)
      continue;
    if ((*p)->data[0]->date < min)
      min = (*p)->data[0]->date;
    for (q = (*p)->data; *q != NULL; q++);
    if ((*(q - 1))->date > max)
      max = (*(q - 1))->date;
    ok = 1;
  }
  if (ok != 1)
    return NULL; /* no data */

  /* allocate room for the data */
  data = calloc((max - min) / INTERVAL + 2, sizeof(stat_feed_t *));
  for (q = data, t = min; t <= max; t += INTERVAL, q++) {
    *q = calloc(1, sizeof(stat_feed_t));
    (*q)->date = t;
  }
  *q = NULL;

  for (p = stat_peers; *p != NULL; p++) {
    if ((*p)->way != way)
      continue;
    for (q = (*p)->data; *q != NULL && (*q)->date < min; q++);
    r = data;
    if ((*q)->date != min)
      for (; *r != NULL && (*r)->date < (*q)->date; r++);
    for (; *r != NULL && *q != NULL; q++, r++) {
      (*r)->offered   += (*q)->offered;
      (*r)->accepted  += (*q)->accepted;
      (*r)->refused   += (*q)->refused;
      (*r)->rejected  += (*q)->rejected;
      (*r)->spooled   += (*q)->spooled;
      (*r)->duplicate += (*q)->duplicate;
      (*r)->missing   += (*q)->missing;
      (*r)->accsize   += (*q)->accsize;
      (*r)->rejsize   += (*q)->rejsize;
      (*r)->count     += (*q)->count;
      (*r)->sec       += (*q)->sec;
    }
  }
  i = (int) time(NULL);
  sprintf(buf, "%s%s/TOTAL-%s-%08X.png", ROOTDIR, TEMPDIR,
	  way == WAY_IN ? "in" : "out", i);
  histo(buf, "Total", data, min_date, max_date, way);
  sprintf(buf, "<A NAME=\"TOTAL-%s\"></A>"
	 "<IMG SRC=\"%s/TOTAL-%s-%08X.png\" WIDTH=%d HEIGHT=%d>\n",
	 way == WAY_IN ? "in" : "out", TEMPDIR, way == WAY_IN ? "in" : "out",
	 i, WIDTH, HEIGHT);

  /* free memory */
  for (q = data; *q != NULL; q++)
    free(*q);
  free(data);
  return strdup(buf);
}

void interpolate(stat_feed_t *cur, stat_peers_t ***peers,
		 time_t tm, int final, int id, char way) {
  int                i;
  stat_peers_t     **peer;
  stat_feed_prev_t **prev;

  /* look for this peer+way in our list */
  for (i = 0, peer = *peers; (*peer != NULL) &&
	 !(streq(f[5 + SKIP_SYSLOG_COLS], (*peer)->name) &&
	   (way == (*peer)->way)); peer++, i++);

  if ((*peer == NULL)) { /* not found. Adding it.. */
    *peers = realloc(*peers, (i + 2) * sizeof(stat_peers_t *));
    for (peer = *peers; *peer != NULL; peer++);
    *peer = (stat_peers_t *) malloc(sizeof(stat_peers_t));
    (*peer)->name = strdup(f[5 + SKIP_SYSLOG_COLS]);
    (*peer)->way = way;
    (*peer)->done = 0;
    (*peer)->data = malloc(sizeof(stat_feed_t *));
    *((*peer)->data) = NULL;
    (*peer)->prevs = malloc(sizeof(stat_feed_prev_t *));
    *((*peer)->prevs) = NULL;
    *(peer + 1) = NULL;
  }
  /* look for previous data for this peer */
  for (i = 0, prev = (*peer)->prevs;
       *prev != NULL && (*prev)->id != id; prev++, i++);

  if (*prev == NULL || (*prev)->id != id) {
    /* it is the first time we see this peer OR
       this peer/id couple. */
    if (final) { /* immediatly use this data */
      compute(&(*peer)->data, NULL, cur);
    }
    else {
      /* new data for this peer/id */
      for (i = 0, prev = (*peer)->prevs;
	   *prev != NULL && (*prev)->id != 0; prev++, i++);
      /* reuse struct with id == 0 (meaning freed) */
      if (*prev == NULL) {
	/* need to extend the structure first */
	(*peer)->prevs = realloc((*peer)->prevs,
				 (i + 2) * sizeof(stat_feed_prev_t *));
	for (prev = (*peer)->prevs;
	     *prev != NULL && (*prev)->id != 0; prev++);
	*prev = (stat_feed_prev_t *)
	  calloc(1, sizeof(stat_feed_prev_t));
	*(prev + 1) = NULL;
      }
      (*prev)->id = id;
      compute(&(*peer)->data, &(*prev)->prev, cur);
    }
  }
  else { /* already have data for this peer/id */
    compute(&(*peer)->data, &(*prev)->prev, cur);
    /* compute results */
    if (final) {
      (*prev)->id            = (*prev)->prev.date      = 0;
      (*prev)->prev.seconds  = (*prev)->prev.offered   = 0;
      (*prev)->prev.accepted = (*prev)->prev.refused   = 0;
      (*prev)->prev.rejected = (*prev)->prev.duplicate = 0;
      (*prev)->prev.spooled  = (*prev)->prev.missing   = 0;
      (*prev)->prev.accsize  = (*prev)->prev.rejsize   = 0;
    }
  }
  /* backup previous data */
  if (!final) {
    (*prev)->prev.date      = tm;
    (*prev)->prev.seconds   = cur->seconds;
    (*prev)->prev.offered   = cur->offered;
    (*prev)->prev.accepted  = cur->accepted;
    (*prev)->prev.refused   = cur->refused;
    (*prev)->prev.rejected  = cur->rejected;
    (*prev)->prev.spooled   = cur->spooled;
    (*prev)->prev.missing   = cur->missing;
    (*prev)->prev.accsize   = cur->accsize;
    (*prev)->prev.rejsize   = cur->rejsize;
    (*prev)->prev.duplicate = cur->duplicate;
  }
}
