/* grapher.c */
/* Copyright (c) 2000-2003 Fabien Tassin <fta@sofaraway.org> */

#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <gd.h>
#include <gdfonts.h>
#include <gdfontt.h>
#include "innpeer-stats.h"

char buffer[BUFSIZE];

char *pretty_size(double d) {
  if      (d > 107374182400.)
    sprintf(buffer, "%.0f GB", d / 1073741824);
  else if (d > 1073741824)
    sprintf(buffer, "%.1f GB", d / 1073741824);
  else if (d > 104857600)
    sprintf(buffer, "%.0f MB", d / 1048576);
  else if (d > 1048576)
    sprintf(buffer, "%.1f MB", d / 1048576);
  else if (d > 102400)
    sprintf(buffer, "%.0f KB", d / 1024);
  else
    sprintf(buffer, "%.1f KB", d / 1024);
  return buffer;
}

void histo(char *filename, char *name, stat_feed_t **data, time_t mindate,
	   time_t maxdate, char way) {
  FILE         *f;
  gdImagePtr    img;
  int		width, height;
  int           white, black, bg, blue, blue2, blue3, red, red2;
  int           green, green2, orange, orange2, purple, purple2;
  int           borderl, borderr, bordert, borderb, first, i, j;
  stat_feed_t **p, **q, o, **old, total;
  ULONG	        xmin, xmax, ox, nx;
  double        ymin, ymax, yzmax, ny, xratio, yratio, yzratio, y;
  char          buf[BUFSIZE];

  borderl = borderr = bordert = borderb = 20;

  img = gdImageCreate(WIDTH, HEIGHT);
  width  = WIDTH;
  height = HEIGHT / 2;

  white   = gdImageColorAllocate(img, 255, 255, 255);
  black   = gdImageColorAllocate(img,   0,   0,   0);
  bg      = gdImageColorAllocate(img, 255, 255, 206);
  blue    = gdImageColorAllocate(img,   0,   0, 255);
  blue2   = gdImageColorAllocate(img,   0,   0, 255);
  blue3   = gdImageColorAllocate(img,  80, 160, 208);
  red     = gdImageColorAllocate(img, 255,   0,   0);
  red2    = gdImageColorAllocate(img, 255,   0,   0);
  green   = gdImageColorAllocate(img,   0, 255,   0);
  green2  = gdImageColorAllocate(img,   0, 255,   0);
  orange  = gdImageColorAllocate(img, 255, 170,   0);
  orange2 = gdImageColorAllocate(img, 255, 170,   0);
  purple  = gdImageColorAllocate(img, 152,  32, 232);
  purple2 = gdImageColorAllocate(img, 152,  32, 232);

  /* by article... */
  xmin = mindate; /* 2147483648UL; */
  if (xmin % INTERVAL != 0)
    xmin += INTERVAL - xmin % INTERVAL;
  else
    xmin += INTERVAL;
  xmax = ymin = ymax = yzmax = 0;
  total.accepted = total.rejected = total.refused = total.missing = 0;
  total.spooled = total.duplicate = total.accsize = total.rejsize = 0;
  for (i = 0, p = data; *p != NULL; p++, i++) {
    /*
    if ((*p)->date < xmin)
      xmin = (*p)->date;
    */
    if ((*p)->date > maxdate)
      break;
    if ((*p)->date > xmax)
      xmax = (*p)->date;
    y = (*p)->accepted + (*p)->refused + (*p)->rejected +
      (*p)->spooled + (*p)->missing + (*p)->duplicate;
    if (y < (*p)->offered)
      y = (*p)->offered;
    if (y > ymax)
      ymax = y;
    y = (*p)->accsize + (*p)->rejsize;
    if (y > yzmax)
      yzmax = y;
    total.accepted  += (*p)->accepted;
    total.refused   += (*p)->refused;
    total.rejected  += (*p)->rejected;
    total.spooled   += (*p)->spooled;
    total.missing   += (*p)->missing;
    total.duplicate += (*p)->duplicate;
    total.accsize   += (*p)->accsize;
    total.rejsize   += (*p)->rejsize;
  }
  if ((int) ymax % YGRAD != 0)
    ymax += YGRAD - (int) ymax % YGRAD;
  sprintf(buf, "%.2f", ymax / INTERVAL);
  j = strlen(buf);
  if ((int) yzmax % YGRAD != 0)
    yzmax += YGRAD - (int) yzmax % YGRAD;
  y = yzmax * 8 / INTERVAL / 1000;
  sprintf(buf, y > 100. ? "%.0f" : "%.1f", y);
  if (j < strlen(buf))
    j = strlen(buf);
  borderl += gdFontSmall->w * j;
  borderr += gdFontTiny->w * strlen("1000000 - 99.9%") + 10;

  old = calloc((i + 1), sizeof(stat_feed_t *));
  xratio  = (xmax - xmin) == 0 ? 1. :
    (double) (width - borderl - borderr - 1) / (xmax - xmin);
  yratio  = (ymax - ymin) == 0 ? 1. :
    (double) (height - borderb - bordert - 1) / (ymax - ymin);
  yzratio = yzmax == 0 ? 1. :
    (double) (height - borderb - bordert - 1) / yzmax;
  xmax += INTERVAL;

  gdImageFilledRectangle(img, 0, 0, WIDTH - 1, HEIGHT - 1, white);
  gdImageRectangle(img, 0, 0, WIDTH - 1, HEIGHT - 1, black);
  gdImageRectangle(img, borderl, bordert - 2, width - borderr - 1,
		   height - borderb - 1, black);
  gdImageRectangle(img, borderl, height + bordert - 2, width - borderr - 1,
		   HEIGHT - borderb - 1, black);

  /* accepted */
  first = 0;
  gdImageFill(img, width / 2, bordert - 1, blue2);
  ox = borderl + (int)(((*data)->date - xmin) * xratio);
  for (p = data, q = old; *p != NULL; p++) {
    if ((*p)->date < xmin) {
      ox = borderl + (int)(((*p)->date - xmin) * xratio);
      continue;
    }
    if ((*p)->date > maxdate)
      break;
    nx = borderl + (int)(((*p)->date - xmin) * xratio);
    ny = height - borderb - 1 - (int)(((*p)->accepted - ymin) * yratio);
    if (first == 0)
      gdImageLine(img, nx,  height - borderb - 1, nx, (int) ny, blue);
    else
      if ((*p)->accepted != 0 || (*(p - 1))->accepted != 0)
	gdImageLine(img, ox, (int) o.accepted, nx, (int) ny, blue);
    o.accepted = ny;
    *q = malloc(sizeof(stat_feed_t));
    (*q)->accepted = ny;
    ox = nx;
    first++;
    q++;
  }
  *q = NULL;

  /* rejected */
  first = 0;
  gdImageFill(img, width / 2, bordert - 1, red2);
  ox = borderl + (int)(((*data)->date - xmin) * xratio);
  for (p = data, q = old; *p != NULL; p++) {
    if ((*p)->date < xmin) {
      ox = borderl + (int)(((*p)->date - xmin) * xratio);
      continue;
    }
    if ((*p)->date > maxdate)
      break;
    nx = borderl + (int)(((*p)->date - xmin) * xratio);
    ny = (*q)->accepted - (int)(((*p)->rejected) * yratio);
    if (first == 0)
      gdImageLine(img, nx,  height - borderb - 1, nx, (int) ny, red);
    else
      if ((*p)->rejected != 0 || (*(p - 1))->rejected != 0)
	gdImageLine(img, ox, (int) o.rejected, nx, (int) ny, red);
    o.rejected = ny;
    (*q)->rejected = ny;
    ox = nx;
    first++;
    q++;
  }

  /* refused */
  first = 0;
  gdImageFill(img, width / 2, bordert - 1, green2);
  ox = borderl + (int)(((*data)->date - xmin) * xratio);
  for (p = data, q = old; *p != NULL; p++) {
    if ((*p)->date < xmin) {
      ox = borderl + (int)(((*p)->date - xmin) * xratio);
      continue;
    }
    if ((*p)->date > maxdate)
      break;
    nx = borderl + (int)(((*p)->date - xmin) * xratio);
    ny = (*q)->rejected - (int)(((*p)->refused) * yratio);
    if (first == 0)
      gdImageLine(img, nx,  height - borderb - 1, nx, (int) ny, green);
    else
      if ((*p)->refused != 0 || (*(p - 1))->refused != 0)
	gdImageLine(img, ox, (int) o.refused, nx, (int) ny, green);
    o.refused = ny;
    (*q)->refused = ny;
    ox = nx;
    first++;
    q++;
  }

  if (way == WAY_IN) {
    /* duplicate */
    first = 0;
    gdImageFill(img, width / 2, bordert - 1, orange2);
    ox = borderl + (int)(((*data)->date - xmin) * xratio);
    for (p = data, q = old; *p != NULL; p++) {
      if ((*p)->date < xmin) {
	ox = borderl + (int)(((*p)->date - xmin) * xratio);
	continue;
      }
      if ((*p)->date > maxdate)
	break;
      nx = borderl + (int)(((*p)->date - xmin) * xratio);
      ny = (*q)->refused - (int)(((*p)->duplicate) * yratio);
      if (first == 0)
	gdImageLine(img, nx,  height - borderb - 1, nx, (int) ny, orange);
      else
	if ((*p)->duplicate != 0 || (*(p - 1))->duplicate != 0)
	  gdImageLine(img, ox, (int) o.duplicate, nx, (int) ny, orange);
      o.duplicate = ny;
      (*q)->duplicate = ny;
      ox = nx;
      first++;
      q++;
    }
    gdImageFill(img, width / 2, bordert - 1, bg);

  }
  else {
    /* missing */
    first = 0;
    gdImageFill(img, width / 2, bordert - 1, orange2);
    ox = borderl + (int)(((*data)->date - xmin) * xratio);
    for (p = data, q = old; *p != NULL; p++) {
      if ((*p)->date < xmin) {
	ox = borderl + (int)(((*p)->date - xmin) * xratio);
	continue;
      }
      if ((*p)->date > maxdate)
	break;
      nx = borderl + (int)(((*p)->date - xmin) * xratio);
      ny = (*q)->refused - (int)(((*p)->missing) * yratio);
      if (first == 0)
	gdImageLine(img, nx,  height - borderb - 1, nx, (int) ny, orange);
      else
	if ((*p)->missing != 0 || (*(p - 1))->missing != 0)
	  gdImageLine(img, ox, (int) o.missing, nx, (int) ny, orange);
      o.missing = ny;
      (*q)->missing = ny;
      ox = nx;
      first++;
      q++;
    }

    /* spooled */
    first = 0;
    gdImageFill(img, width / 2, bordert - 1, purple2);
    ox = borderl + (int)(((*data)->date - xmin) * xratio);
    for (p = data, q = old; *p != NULL; p++) {
      if ((*p)->date < xmin) {
	ox = borderl + (int)(((*p)->date - xmin) * xratio);
	continue;
      }
      if ((*p)->date > maxdate)
	break;
      nx = borderl + (int)(((*p)->date - xmin) * xratio);
      ny = (*q)->missing - (int)(((*p)->spooled) * yratio);
      if (first == 0)
	gdImageLine(img, nx,  height - borderb - 1, nx, (int) ny, purple);
      else
	if ((*p)->spooled != 0 || (*(p - 1))->spooled != 0)
	  gdImageLine(img, ox, (int) o.spooled, nx, (int) ny, purple);
      o.spooled = ny;
      (*q)->spooled = ny;
      ox = nx;
      first++;
      q++;
    }
    gdImageFill(img, width / 2, bordert - 1, bg);
  }

  /* accepted size */
  first = 0;
  gdImageFill(img, width / 2, height + bordert - 1, blue2);
  ox = borderl + (int)(((*data)->date - xmin) * xratio);
  for (p = data, q = old; *p != NULL; p++) {
    if ((*p)->date < xmin) {
      ox = borderl + (int)(((*p)->date - xmin) * xratio);
      continue;
    }
    if ((*p)->date > maxdate)
      break;
    nx = borderl + (int)(((*p)->date - xmin) * xratio);
    ny = HEIGHT - borderb - 1 - (int)(((*p)->accsize) * yzratio);
    if (first == 0)
      gdImageLine(img, nx,  HEIGHT - borderb - 1, nx, (int) ny, blue);
    else
      if ((*p)->accsize != 0 || (*(p - 1))->accsize != 0)
	gdImageLine(img, ox, (int) o.accsize, nx, (int) ny, blue);
    o.accsize = ny;
    (*q)->accsize = ny;
    ox = nx;
    first++;
    q++;
  }

  /* rejected size */
  first = 0;
  gdImageFill(img, width / 2, height + bordert - 1,
	      way == WAY_IN ? orange2 : red2);
  ox = borderl + (int)(((*data)->date - xmin) * xratio);
  for (p = data, q = old; *p != NULL; p++) {
    if ((*p)->date < xmin) {
      ox = borderl + (int)(((*p)->date - xmin) * xratio);
      continue;
    }
    if ((*p)->date > maxdate)
      break;
    nx = borderl + (int)(((*p)->date - xmin) * xratio);
    ny = (*q)->accsize - (int)(((*p)->rejsize) * yzratio);
    if (first == 0)
      gdImageLine(img, nx,  HEIGHT - borderb - 1, nx, (int) ny,
      		  way == WAY_IN ? orange : red);
    else
      if ((*p)->rejsize != 0 || (*(p - 1))->rejsize != 0)
	gdImageLine(img, ox, (int) o.rejsize, nx, (int) ny,
		    way == WAY_IN ? orange : red);
    o.rejsize = ny;
    (*q)->rejsize = ny;
    ox = nx;
    first++;
    q++;
  }
  gdImageFill(img, width / 2, height + bordert - 1, bg);

  /* complete the graphs */
  gdImageFilledRectangle(img, borderl - 1, bordert - 2, width - borderr,
			 bordert, white);
  gdImageRectangle(img, borderl, bordert, width - borderr - 1,
		   height - borderb - 1, black);
  gdImageFilledRectangle(img, borderl - 1, height + bordert - 2,
			 width - borderr, height + bordert, white);
  gdImageRectangle(img, borderl, height + bordert, width - borderr - 1,
		   HEIGHT - borderb - 1, black);

  /* add graduations to each graph */
  {
    int styleDotted[4];
    struct tm *tm;

    styleDotted[0] = styleDotted[1] = styleDotted[2] = gdTransparent;
    styleDotted[3] = black;

    if (ymax == 0)
      ymax = INTERVAL;
    if (yzmax == 0)
      yzmax = INTERVAL * 1000;

    gdImageSetStyle(img, styleDotted, 4);
    for (nx = xmin - (xmin % 3600) + ((xmin % 3600 == 0) ? 0 : 3600);
	 nx < xmax; nx += 3600) {
      gdImageLine(img, borderl + (int) ((nx - xmin) * xratio), bordert,
		  borderl + (int) ((nx - xmin) * xratio),
		  height - borderb - 1, gdStyled);
      gdImageLine(img, borderl + (int) ((nx - xmin) * xratio),
		  height + bordert, borderl + (int) ((nx - xmin) * xratio),
		  HEIGHT - borderb - 1, gdStyled);
      if (nx < xmax - 1600) {
	tm = localtime((const time_t *) &nx);
	sprintf(buf, "%02d", tm->tm_hour);
	gdImageString(img, gdFontSmall,
		      borderl + (int) ((nx - xmin + 1800) * xratio) -
		      gdFontSmall->w + 2, height - borderb + 2, (UCHAR *) buf,
		      black);
	gdImageString(img, gdFontSmall,
		      borderl + (int) ((nx - xmin + 1800) * xratio) -
		      gdFontSmall->w + 2, HEIGHT - borderb + 2, (UCHAR *) buf,
		      black);
      }
    }
    for (i = 0; i <= YGRAD; i++) {
      ny = height - borderb - 1 - (ymax - ymin - 1) * yratio * i / YGRAD;
      gdImageLine(img, borderl - 3,
		  (int) ny, width - borderr, (int) ny, gdStyled);
      sprintf(buf, "%.2f", ymax * i / YGRAD / INTERVAL);
      gdImageString(img, gdFontSmall,
		    borderl - 3 - strlen(buf) * gdFontSmall->w,
		    (int) ny - gdFontSmall->h / 2, (UCHAR *) buf, black);
      /* size */
      ny = HEIGHT - borderb - 1 - (yzmax - 1) * yzratio * i / YGRAD;
      gdImageLine(img, borderl - 3,
		  (int) ny, width - borderr, (int) ny, gdStyled);
      y = yzmax * i / YGRAD * 8 / INTERVAL / 1000;
      sprintf(buf, y > 100. ? "%.0f" : "%.1f", y);
      gdImageString(img, gdFontSmall,
		    borderl - 3 - strlen(buf) * gdFontSmall->w,
		    (int) ny - gdFontSmall->h / 2, (UCHAR *) buf, black);
    }
    gdImageStringUp(img, gdFontSmall, 2, (height + gdFontSmall->w * 7) / 2,
                    (UCHAR *) "art/sec", black);
    gdImageStringUp(img, gdFontSmall, 2, (3 * height + gdFontSmall->w * 7) / 2,
		    (UCHAR *) "Kb/sec", black);
  }

  /* legend */
  {
    int y = bordert;
    double tot = total.accepted + total.rejected + total.refused +
      total.missing + total.spooled + total.duplicate;

    if (tot == 0)
      tot = 1; /* to avoid division by zero */
    gdImageString(img, gdFontSmall, width - borderr + 22, y - 2,
		  (UCHAR *) "accepted", black);
    sprintf(buf, "%d - %.1f%%", (int) total.accepted,
	    total.accepted / tot * 100);
    gdImageString(img, gdFontTiny, width - borderr + 22, y + 10,
		  (UCHAR *) buf, black);
    gdImageFilledRectangle(img, width - borderr + 6, y, width - borderr + 16,
			   y + 10, blue);
    gdImageRectangle(img, width - borderr + 6, y, width - borderr + 16, y + 10,
		     black);
    y += 27;
    gdImageString(img, gdFontSmall, width - borderr + 22, y - 2,
		  (UCHAR *) "rejected", black);
    sprintf(buf, "%d - %.1f%%", (int) total.rejected,
	    total.rejected / tot * 100);
    gdImageString(img, gdFontTiny, width - borderr + 22, y + 10,
		  (UCHAR *) buf, black);
    gdImageFilledRectangle(img, width - borderr + 6, y, width - borderr + 16,
			   y + 10, red);
    gdImageRectangle(img, width - borderr + 6, y, width - borderr + 16, y + 10,
		     black);
    y += 27;
    gdImageString(img, gdFontSmall, width - borderr + 22, y - 2,
		  (UCHAR *) "refused", black);
    sprintf(buf, "%d - %.1f%%", (int) total.refused,
	    total.refused / tot * 100);
    gdImageString(img, gdFontTiny, width - borderr + 22, y + 10,
		  (UCHAR *) buf, black);
    gdImageFilledRectangle(img, width - borderr + 6, y, width - borderr + 16,
			   y + 10, green);
    gdImageRectangle(img, width - borderr + 6, y, width - borderr + 16, y + 10,
		     black);
    y += 27;
    if (way == WAY_IN) {
      gdImageString(img, gdFontSmall, width - borderr + 22, y - 2,
		    (UCHAR *) "duplicate", black);
      sprintf(buf, "%d - %.1f%%", (int) total.duplicate,
	      total.duplicate / tot * 100);
      gdImageString(img, gdFontTiny, width - borderr + 22, y + 10,
		    (UCHAR *) buf, black);
      gdImageFilledRectangle(img, width - borderr + 6, y, width - borderr + 16,
			     y + 10, orange);
      gdImageRectangle(img, width - borderr + 6, y, width - borderr + 16,
		       y + 10, black);
    }
    else {
      gdImageString(img, gdFontSmall, width - borderr + 22, y - 2,
		    (UCHAR *) "missing", black);
      sprintf(buf, "%d - %.1f%%", (int) total.missing,
	      total.missing / tot * 100);
      gdImageString(img, gdFontTiny, width - borderr + 22, y + 10,
		    (UCHAR *) buf, black);
      gdImageFilledRectangle(img, width - borderr + 6, y, width - borderr + 16,
			     y + 10, orange);
      gdImageRectangle(img, width - borderr + 6, y, width - borderr + 16,
		       y + 10, black);
      y += 27;
      gdImageString(img, gdFontSmall, width - borderr + 22, y - 2,
		    (UCHAR *) "spooled", black);
      sprintf(buf, "%d - %.1f%%", (int) total.spooled,
	      total.spooled / tot * 100);
      gdImageString(img, gdFontTiny, width - borderr + 22, y + 10,
		    (UCHAR *) buf, black);
      gdImageFilledRectangle(img, width - borderr + 6, y, width - borderr + 16,
			     y + 10, purple);
      gdImageRectangle(img, width - borderr + 6, y, width - borderr + 16,
		       y + 10, black);
    }
    /* by size */
    y = height + bordert;
    tot = total.accsize + total.rejsize;
    if (tot == 0)
      tot = 1; /* to avoid division by zero */

    gdImageString(img, gdFontSmall, width - borderr + 22, y - 2,
		  (UCHAR *) "accepted", black);
    sprintf(buf, "%s - %.1f%%", pretty_size(total.accsize),
	    total.accsize / tot * 100);
    gdImageString(img, gdFontTiny, width - borderr + 22, y + 10,
		  (UCHAR *) buf, black);
    gdImageFilledRectangle(img, width - borderr + 6, y, width - borderr + 16,
			   y + 10, blue);
    gdImageRectangle(img, width - borderr + 6, y, width - borderr + 16, y + 10,
		     black);
    y += 27;
    gdImageString(img, gdFontSmall, width - borderr + 22, y - 2,
		  (UCHAR *) (way == WAY_IN ? "duplicate" : "rejected"), black);
    sprintf(buf, "%s - %.1f%%", pretty_size(total.rejsize),
	    total.rejsize / tot * 100);
    gdImageString(img, gdFontTiny, width - borderr + 22, y + 10,
		  (UCHAR *) buf, black);
    gdImageFilledRectangle(img, width - borderr + 6, y, width - borderr + 16,
			   y + 10, way == WAY_IN ? orange : red);
    gdImageRectangle(img, width - borderr + 6, y, width - borderr + 16, y + 10,
		     black);
  }
  /* peer name */
  sprintf(buf, "%s: %s (by article)", name,
	  way == WAY_IN ? "Incoming" : "Outgoing");
  gdImageString(img, gdFontSmall, borderl + 4, 4, (UCHAR *) buf, black);
  sprintf(buf, "%s: %s (by size)", name,
	  way == WAY_IN ? "Incoming" : "Outgoing");
  gdImageString(img, gdFontSmall, borderl + 4, height + 4, (UCHAR *) buf,
		black);

  for (q = old; *q != NULL; q++)
    free(*q);
  free(old);

  /* write the graph into a file */
  if ((f = fopen(filename, "wb")) == NULL) {
    printf("Can't open '%s': %s\n", filename, strerror(errno));
    exit(1);
  }
  gdImagePng(img, f);
  fclose(f);
  gdImageDestroy(img);
}
