/* groups.c */
/* Copyright (c) 2000-2003 Fabien Tassin <fta@sofaraway.org> */

#include <errno.h>
#include <string.h>
#include <stdio.h>
#include "innpeer-stats.h"

int streq(char *s1, char *s2) {
  return (strncmp(s1, s2, strlen(s1)) == 0 && strlen(s1) == strlen(s2));
}

stat_group_t **parse_groups_conf(char *filename, stat_peers_t **stat_peers) {
  stat_group_t **groups, **p;
  FILE          *F;
  FIO           *fio;
  char          *s, *r, **q;
  char         **elts = NULL;
  int		 num = 0, ng = 0, nm, alloc;
  stat_peers_t **peer, **t;

  groups = (stat_group_t **) malloc(sizeof(stat_group_t *));
  *groups = NULL;
  if (filename != NULL) {
    if ((F = fopen(filename, "r")) == NULL) {
      fprintf(stdout, "Can't open file %s: %s\n", filename, strerror(errno));
      exit(0);
    }
    fio = fio_init(fileno(F));
    for (;;) {
      s = fio_readline(fio);
      num++;
      if (s == NULL)
	break;
      if ((r = strchr(s, '#')) != NULL)
	*r = '\0';
      for (; *s == ' ' || *s == '\t'; s++);
      if (*s == '\0')
	continue;
      split(s, &elts);
      /* got the name of the group */
      if ((*elts)[strlen(*elts) - 1] != ':') {
	fprintf(stderr, "Syntax error in %s line %d: malformed line\n",
		filename, num);
	exit(0);
      }
      (*elts)[strlen(*elts) - 1] = '\0';
      alloc = 0;
      nm = 0;
      p = NULL;
      /* look for associations */
      for (peer = stat_peers; *peer != NULL; peer++) {
	for (q = elts + 1; *q != NULL; q++) {
	  if (streq((*peer)->name, *q)) {
	    if (alloc == 0) {
	      alloc++;
	      ng++;
	      groups = realloc(groups, (ng + 1) * sizeof(stat_group_t *));
	      for (p = groups; *p != NULL; p++);
	      *p = malloc(sizeof(stat_group_t));
	      *(p + 1) = NULL;
	      (*p)->name = strdup(*elts);
	      (*p)->members = (stat_peers_t **) malloc(sizeof(stat_peers_t *));
	      *(*p)->members = NULL;
	      (*p)->nmembers = 0;
	    }
	    (*p)->nmembers++;
	    (*p)->members = realloc((*p)->members,
			      ((*p)->nmembers + 1) * sizeof(stat_peers_t *));
	    for (t = (*p)->members; *t != NULL; t++);
	    *t = *peer;
	    /* mark it as done */
	    (*t)->done++;
	    *(t + 1) = NULL;
	    (*t)->in_group = *p;
	  }
	}
      }
    }
    free(*elts);
    free(elts);
    fio_free(fio);
    fclose(F);
  }
  /* create a group for each unused (ungrouped) peer */
  for (peer = stat_peers; *peer != NULL; peer++) {
    if ((*peer)->done != 0)
      continue;
    (*peer)->done++;
    /* look for an already existing group with the same name.. */
    for (p = groups; *p != NULL && !streq((*p)->name, (*peer)->name); p++);
    if (*p == NULL) { /* not found */
      ng++;
      groups = realloc(groups, (ng + 1) * sizeof(stat_group_t *));
      for (p = groups; *p != NULL; p++);
      *(p + 1) = NULL;
      *p = malloc(sizeof(stat_group_t));
      (*p)->name = strdup((*peer)->name);
      (*p)->members = (stat_peers_t **) malloc(2 * sizeof(stat_peers_t *));
      (*p)->members[0] = *peer;
      (*p)->members[1] = NULL;
      (*peer)->in_group = *p;
      (*p)->nmembers = 1;
    }
    else { /* found */
      (*p)->members = realloc((*p)->members,
			      ((*p)->nmembers + 2) * sizeof(stat_peers_t *));
      (*p)->members[(*p)->nmembers] = *peer;
      (*peer)->in_group = *p;
      (*p)->nmembers++;
      (*p)->members[(*p)->nmembers] = NULL;
    }
  }
  return groups;
}

void free_groups(stat_group_t **groups) {
  stat_group_t **g;

  for (g = groups; *g != NULL; g++) {
    if ((*g)->name != NULL)
      free((*g)->name);
    if ((*g)->members != NULL)
      free((*g)->members);
    free(*g);
  }
  free(groups);
}
