/* innpeer-stats.h */
/* Copyright (c) 2000-2003 Fabien Tassin <fta@sofaraway.org> */

#ifndef _INNPEER_STATS_H_
#define _INNPEER_STATS_H_

/* Skip columns from syslog output (often useful for Solaris) */
/* "date hostname process[pid]: foo" vs "date hostname process[pid]: [ID num loglevel] foo" */
#define SKIP_SYSLOG_COLS	0

#define LOGFILE		"/var/log/news/news.notice"
#define ROOTDIR		"./var/www"		/* no trailing slash (nts) */
#define DATADIR         "/innpeer-stats"    /* as seen from Web server + nts */
#define TEMPDIR		"/innpeer-stats/tmp"/* as seen from Web server + nts */

/* various default values (for CGI) that can be changed on command line */
#define GROUPRULESFILE  "/news/etc/innpeer-groups.conf"
#define SPLIT_IN_FILES  0   /* boolean */
#define GROUP_PEERS     0   /* boolean */

#define HTML_HEADER	"/news/etc/innpeer-header.html"
#define HTML_FOOTER	"/news/etc/innpeer-footer.html"
#define CHANGE_HEADERS  0   /* If true, the previous 2 files must exist */

#define COLS_IN         3   /* list incoming peers in that number of columns */
#define COLS_OUT        5   /* list outgoing peers in that number of columns */

#define WIDTH		600 /* graph width in pixels */
#define HEIGHT		350 /* graph height in pixels */
#define YGRAD		5   /* number of vertical graduations */

/* should be handled by autoconf */
#define HAVE_MCHECK_H   0

/* ********************************************************************** */

#include <time.h>
#include <stdlib.h>
#include "fio.h"

#ifdef MALLOC
# include <dmalloc.h>
#endif

#if HAVE_MCHECK_H == 1
# ifdef MTRACE
#  include <mcheck.h>
# endif /* MTRACE */
#else
# undef MTRACE
#endif /* HAVE_MCHECK_H */

#define PRGNAME		"innpeer-stats"
#define VERSION		"0.03"
#define LOCATION        "http://www.sofaraway.org/projects/innpeer-stats/"

#define BUFSIZE		255

#define INTERVAL	300 /* 5 minutes */

#define APP_INND	0x01
#define APP_INNFEED	0x02

#define WAY_IN		0x01
#define WAY_OUT		0x02

typedef unsigned long ULONG;
typedef unsigned char UCHAR;

typedef struct _stat_peers_t stat_peers_t;

typedef struct _stat_group_t {
  char         *name;
  stat_peers_t **members;
  int           nmembers;
} stat_group_t;

typedef struct _stat_queue_t {
  double length;
  double max;
  double empty;
  double quartile0;
  double quartile25;
  double quartile50;
  double quartile75;
  double full;
} stat_queue_t;

typedef struct _stat_feed_t {
  time_t        date;
  ULONG	        seconds;
  double        offered;
  double        accepted;
  double        refused;
  double        rejected;
  double        spooled;
  double        missing;
  double        duplicate;
  double        accsize;
  double        rejsize;
  stat_queue_t  queue; /* only used for innfeed stats */
  double        count;
  double        sec;
} stat_feed_t;

typedef struct _stat_feed_prev_t {
  stat_feed_t prev;
  int         id;
} stat_feed_prev_t;

struct _stat_peers_t {
  char              *name;
  char		     way;
  stat_feed_t      **data;
  stat_feed_prev_t **prevs;
  char               done;
  stat_group_t      *in_group;
};

/* parser.c */
int   parse(FIO *fio, stat_peers_t ***peers);
void  free_data(stat_peers_t **data);
int   split(char *buf, char ***tab);

/* compute.c */
void  interpolate(stat_feed_t *cur, stat_peers_t ***peers,
		  time_t tm, int final, int id, char way);
void  compute(stat_feed_t ***data, stat_feed_t *old, stat_feed_t *new);
char *compute_total(stat_peers_t **stat_peers, char way);

/* groups.c */
int            streq(char *s1, char *s2);
stat_group_t **parse_groups_conf(char *filename, stat_peers_t **stat_peers);
void           free_groups(stat_group_t **groups);

/* grapher.c */
void  histo(char *filename, char *name, stat_feed_t **data, time_t mindate,
	    time_t maxdate, char way);

#endif /* _INNPEER_STATS_H_ */
