/* main.c */
/* Copyright (c) 2000-2003 Fabien Tassin <fta@sofaraway.org> */

#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>
#include "innpeer-stats.h"

time_t max_date, min_date = 0;
char  **f = NULL;

int isort(const void *va, const void *vb) {
  stat_peers_t **a, **b;

  a = (stat_peers_t **)va;
  b = (stat_peers_t **)vb;
  return strcasecmp((*a)->name, (*b)->name);
}

void print_header(FILE *F, char *name, char *file) {
  FILE *G;
  char buf[BUFSIZE];
  size_t s;

  fprintf(F, "<!-- %s version %s (c) 2000-2003 by Fabien Tassin -->\n"
	  "<!-- Created by Fabien Tassin -+- fta@sofaraway.org -->\n"
	  "<HTML>\n<HEAD><TITLE>INN Peer Statistics : %s</TITLE></HEAD>\n",
	  PRGNAME, VERSION, name == NULL ? "index" : name);
  if (file != NULL) {
    fprintf(F, "<!-- begin %s -->\n", file);
    if ((G = fopen(file, "r")) == NULL)
      fprintf(F, "Can't open file %s: %s\n", file, strerror(errno));
    else {
      while (!feof(G)) {
	s = fread(buf, 1, BUFSIZE, G);
	buf[s] = '\0';
	fprintf(F, buf);
      }
      fclose(G);
    }
    fprintf(F, "<!-- end %s -->\n", file);
  }
  else
    fprintf(F, "<BODY BGCOLOR=\"#FFFFFF\">\n");
  fprintf(F, "<CENTER><FONT SIZE=\"+3\">"
	  "<B>INN Peer Statistics%s%s</B></FONT>\n<P>%s - ",
	  name == NULL ? "" : " : ", name == NULL ? "" : name,
	  ctime(&min_date));
  fprintf(F, "%s</CENTER>\n", ctime(&max_date));

}

void print_footer(FILE *F, char *file) {
  FILE *G;
  char buf[BUFSIZE];
  size_t s;

  fprintf(F, "<P>\n<HR NOSHADE>\n<A HREF=\"%s\">%s</A> %s "
	  "&copy; Copyright 2000-2003 Fabien Tassin\n", LOCATION,
	  PRGNAME, VERSION);
  if (file != NULL) {
    fprintf(F, "<!-- begin %s -->\n", file);
    if ((G = fopen(file, "r")) == NULL)
      fprintf(F, "Can't open file %s: %s\n", file, strerror(errno));
    else {
      while (!feof(G)) {
	s = fread(buf, 1, BUFSIZE, G);
	buf[s] = '\0';
	fprintf(F, buf);
      }
      fclose(G);
    }
    fprintf(F, "<!-- end %s -->\n", file);
  }
  fprintf(F, "</BODY>\n</HTML>\n");
}

int main(int argc, char *argv[]) {
  FILE          *F, *G, *H;
  FIO           *fio;
  char           buf[BUFSIZE], *s;
  stat_peers_t **p, **stat_peers = NULL;
  int            i, num, t, nin;
  int            not_a_cgi = 0;
  int            usage = 0;
  int            split_in_files = SPLIT_IN_FILES;
  char          *group_filename = NULL;
  char          *file_header = NULL, *file_footer = NULL;
  stat_group_t **groups, **g;

#ifdef MTRACE
  mtrace();
#endif /* MTRACE */

  if (GROUP_PEERS) {
    group_filename = malloc(sizeof(GROUPRULESFILE) + 1);
    strcpy(group_filename, GROUPRULESFILE);
    split_in_files = 1;
  }

  if (CHANGE_HEADERS) {
    file_header = malloc(sizeof(HTML_HEADER) + 1);
    strcpy(file_header, HTML_HEADER);
    file_footer = malloc(sizeof(HTML_FOOTER) + 1);
    strcpy(file_footer, HTML_FOOTER);
  }

  /* parse command line */
  while (1) {
    int c = -1;

    opterr = 0; /* prevent error messages */
    c = getopt(argc, argv, "ChvfgG:H:F:");
    if (c == -1)
      break;
    switch(c) {
    case 'f':
      split_in_files = 1;
      break;
    case 'g':
      if (group_filename != NULL)
	free(group_filename);
      group_filename = malloc(sizeof(GROUPRULESFILE) + 1);
      strcpy(group_filename, GROUPRULESFILE);
      split_in_files = 1;
      break;
    case 'G':
      if (group_filename != NULL)
	free(group_filename);
      group_filename = strdup(optarg);
      split_in_files = 1;
      break;
    case 'C':
      not_a_cgi = 1;
      break;
    case 'H':
      if (file_header != NULL)
	free(file_header);
      file_header = strdup(optarg);
      break;
    case 'F':
      if (file_footer != NULL)
	free(file_footer);
      file_footer = strdup(optarg);
      break;
    case 'h':
      usage = 1;
      break;
    case 'v':
      printf("%s version %s\n", PRGNAME, VERSION);
      exit(0);
    default:
      fprintf(stderr, "invalid option '%c'\n", optopt);
      usage = 1;
    }
  }

  if (usage == 1) {
    printf("Usage: %s [-hv] [-Cfg] [-[GHF] filename] [logfile]\n"
	   "-h  show this usage message.\n"
	   "-v  show version.\n"
	   "-C  do not run as a CGI.\n"
	   "-H  use this HTML file as header.\n"
	   "-F  use this HTML file as footer.\n"
	   "-f  generate one file per peer.\n"
	   "-g  group peers using default rules file [%s].\n"
	   "-G  group peers using this rules file.\n",
	   PRGNAME, GROUPRULESFILE);
    if (group_filename != NULL)
      free(group_filename);
    exit(0);
  }

  if (not_a_cgi == 0)
    printf("Content-type: text/html\n\n");
    /* printf("Content-type: text/html\nPragma: no-cache\n\n"); */

  /* As a CGI, if we've decided to split the results by peers, we should
     print the index page to stdout and everything else to a filehandle.
     Else (no split), only the graphs go to a fh, all the text will go
     to stdout. If the user requires us to run not as a CGI, everything
     will go to a fh, splited or not. */

  strcpy(buf, argc > optind ? argv[optind] : LOGFILE);
  if ((F = fopen(buf, "r")) == NULL) {
    fprintf(not_a_cgi ? stderr : stdout,
	    "Can't open file %s: %s\n", buf, strerror(errno));
    if (group_filename != NULL)
      free(group_filename);
    exit(1);
  }
  stat_peers = (stat_peers_t **) malloc(sizeof(stat_peers_t *));
  *stat_peers = NULL;
  fio = fio_init(fileno(F));
  num = parse(fio, &stat_peers);
  free(*f);
  free(f);
  fio_free(fio);
  fclose(F);
  qsort(stat_peers, num, sizeof(*stat_peers), isort);
  if (split_in_files || group_filename)
    groups = parse_groups_conf(group_filename, stat_peers);
  else {
    groups = (stat_group_t **) malloc(2 * sizeof(stat_group_t *));
    *groups = malloc(sizeof(stat_group_t));
    (*groups)->name = NULL; /* unlabelled group */
    (*groups)->members = (stat_peers_t **)
      malloc((num + 1) * sizeof(stat_peers_t *));
    for (p = stat_peers, i = 0; *p != NULL; p++, i++) {
      (*groups)->members[i] = *p;
      (*p)->in_group = *groups;
    }
    (*groups)->nmembers = num;
    (*groups)->members[i] = NULL;
    groups[1] = NULL;
  }

  /* count incoming peers */
  for (p = stat_peers, nin = 0; *p != NULL; p++)
    if ((*p)->way == WAY_IN)
      nin++;

  /* get a filehandle */
  if (not_a_cgi) {
    sprintf(buf, "%s%s/index.html", ROOTDIR, DATADIR);
    if ((G = fopen(buf, "w")) == NULL) {
    fprintf(not_a_cgi ? stderr : stdout,
	    "Can't create file %s: %s\n", buf, strerror(errno));
      if (group_filename != NULL)
	free(group_filename);
      exit(1);
    }
  }
  else {
    G = stdout;
  }

  /* display the results */
  print_header(G, NULL, file_header);
  if ((s = compute_total(stat_peers, WAY_IN)) != NULL) {
    fprintf(G, "<HR NOSHADE><P>\n<H1>Incoming statistics</H1>\n<P>\b"
	    "<B>Global :</B><P><CENTER>\n"
	    "\n%s</CENTER>\n<P><P>\n<B>By Peer :</B><P>\n"
	    "<TABLE BORDER=0>\n <TR>\n  <TD VALIGN=TOP>\n   <UL>\n", s);
    free(s);
    t = nin > COLS_IN ? nin / COLS_IN + (nin % COLS_IN != 0) : nin;
    for (p = stat_peers, i = 0; *p != NULL; p++) {
      if ((*p)->way != WAY_IN)
	continue;
      if ((i % t == 0) && (i != 0))
	fprintf(G, "   </UL>\n  </TD>\n  <TD VALIGN=TOP>\n   <UL>\n");
      sprintf(buf, "%s/%s.html", not_a_cgi ? DATADIR : TEMPDIR,
	      (*p)->in_group->name);
      fprintf(G, "    <LI><A HREF=\"%s#%s-in\">%s</A>\n",
	      split_in_files ? buf : "", (*p)->name, (*p)->name);
      i++;
    }
    fprintf(G, "   </UL>\n  </TD>\n </TR>\n</TABLE>\n<P>\n");
  }
  if ((s = compute_total(stat_peers, WAY_OUT)) != NULL) {
    fprintf(G, "<HR NOSHADE><P>\n<H1>Outgoing statistics</H1>\n<P>\n"
	    "<B>Global :</B><P><CENTER>\n"
	    "%s</CENTER>\n<P>\n<B>By Peer :</B><P>\n"
	    "<TABLE BORDER=0><TR><TD VALIGN=TOP><UL>\n", s);
    free(s);
    t = (num - nin) > COLS_OUT ?
      (num - nin) / COLS_OUT + ((num - nin) % COLS_OUT != 0) : (num - nin);
    for (p = stat_peers, i = 0; *p != NULL; p++) {
      if ((*p)->way != WAY_OUT)
	continue;
      if ((i % t == 0) && (i != 0))
	fprintf(G, "   </UL>\n  </TD>\n  <TD VALIGN=TOP>\n   <UL>\n");
      sprintf(buf, "%s/%s.html", not_a_cgi ? DATADIR : TEMPDIR,
	      (*p)->in_group->name);
      fprintf(G, "    <LI><A HREF=\"%s#%s-out\">%s</A>\n",
	      split_in_files ? buf : "", (*p)->name, (*p)->name);
      i++;
    }
    fprintf(G, "   </UL>\n  </TD>\n </TR>\n</TABLE>\n<P>\n");
  }

  i = (int) time(NULL);
  for (g = groups; *g != NULL; g++) {
    char nway[WAY_OUT + 1];
    int  way;

    nway[WAY_IN] = nway[WAY_OUT] = 0;
    /* determine if there is at least one IN (resp. OUT) feed */
    for (p = (*g)->members; *p != NULL; p++) {
      if ((*p)->way == WAY_IN)
	nway[WAY_IN]  = 1;
      else
	nway[WAY_OUT] = 1;
    }
    if (split_in_files == 1) {
      sprintf(buf, "%s%s/%s.html", ROOTDIR, not_a_cgi ? DATADIR : TEMPDIR,
	      (*g)->name == NULL ? "all" : (*g)->name);
      if ((H = fopen(buf, "w")) == NULL) {
	fprintf(not_a_cgi ? stderr : stdout,
		"Can't create file %s: %s\n", buf, strerror(errno));
	exit(1);
      }
      print_header(H, (*g)->name, file_header);
      fprintf(H, "<P>\n");
    }
    else
      H = G;
    for (way = WAY_IN; way <= WAY_OUT; way++) {
      if (nway[way] == 1)
	fprintf(H, "<HR NOSHADE><P>\n<H1>%s statistics%s</H1>\n<P>\n"
		"<CENTER>\n", way == WAY_IN ? "Incoming" : "Outgoing",
		split_in_files == 1 ? "" : " (by peer)");
      for (p = (*g)->members; *p != NULL; p++) {
	if ((*p)->way != way)
	  continue;
	sprintf(buf, "%s%s/%s-%s-%08X.png", ROOTDIR, TEMPDIR, (*p)->name,
		way == WAY_IN ? "in" : "out", i);
	histo(buf, (*p)->name, (*p)->data, min_date, max_date, way);
	fprintf(H, "<A NAME=\"%s-%s\"></A>"
		"<IMG SRC=\"%s/%s-%s-%08X.png\" WIDTH=%d HEIGHT=%d><P>\n",
		(*p)->name, way == WAY_IN ? "in" : "out", TEMPDIR, (*p)->name,
		way == WAY_IN ? "in" : "out", i, WIDTH, HEIGHT);
      }
      if (nway[way] == 1)
	fprintf(H, "</CENTER>\n");
    }
    if (split_in_files == 1) {
      print_footer(H, file_footer);
      fclose(H);
    }
  }
  print_footer(G, file_footer);
  if (not_a_cgi)
    fclose(G);
  free_data(stat_peers);
  free_groups(groups);
  if (group_filename != NULL)
    free(group_filename);
  return(0);
}
