/* parser.c */
/* Copyright (c) 2000-2003 Fabien Tassin <fta@sofaraway.org> */

#include <string.h>
#include "innpeer-stats.h"

static char     MONTHS[] = "JanFebMarAprMayJunJulAugSepOctNovDec";
extern char   **f;
extern time_t   max_date, min_date;

int split(char *buf, char ***tab) {
  char **t, *p;
  int    i;

  /* free previously allocated memory. *tab should be initialized to NULL
     the first time */
  if (*tab != NULL) {
    free(*tab[0]);
    free(*tab);
  }
  /* skip separators */
  while (*buf == ' ' || *buf == '\t')
    buf++;
  i = strlen(buf);
  p = strdup(buf);

  /* split by separators */
  for (*tab = t = (char **) malloc((i + 2) * sizeof(char *)); *p != '\0';) {
    for (*t++ = p; *p != '\0' && *p != ' ' && *p != '\t';)
      p++;
    if (*p == '\0')
      break;
    for (*p++ = '\0'; *p == ' ' || *p == '\t';)
      p++;
  }
  *t = NULL;
  /* return number of parts */
  return(t - *tab);
}

void parse_innfeed(stat_peers_t ***peers, time_t tm, char **f) {
  char              *p;
  int                id, i;
  stat_feed_t        cur;
  char              *pos[8];

  /* get the pid of the process producing this log line */
  id = 0;
  if ((p = strchr(f[4], ']')) != NULL) {
    *p = '\0';
    p = strchr(f[4], '[');
    id = atoi(p + 1);
  }
  /* process innfeed 'checkpoint' and 'final' lines */
  if (streq(f[6 + SKIP_SYSLOG_COLS], "checkpoint") ||
      streq(f[6 + SKIP_SYSLOG_COLS], "final")) {
    int final = 0;

    if (f[6 + SKIP_SYSLOG_COLS][0] == 'f')
      final = 1;
    if (strchr(f[5 + SKIP_SYSLOG_COLS], ':') != NULL) {
      /* A sub-connection */
      /* Ignored. Can be processed later */
    }
    else {
      cur.date      = tm;
      cur.seconds   = atol(f[8 + SKIP_SYSLOG_COLS]);
      cur.duplicate = 0;
      if (streq(f[11 + SKIP_SYSLOG_COLS], "on_close")) {
	/* final seconds 30 spooled 73 on_close 15 sleeping 58 */
	cur.offered  = cur.accepted = cur.refused = cur.rejected = 0.;
	cur.missing  = cur.accsize  = cur.rejsize = 0.;
	cur.spooled  = atof(f[10 + SKIP_SYSLOG_COLS]);
	cur.queue.length     = cur.queue.max        = cur.queue.empty = 0.;
	cur.queue.quartile0  = cur.queue.quartile25 = 0;
	cur.queue.quartile50 = cur.queue.quartile75 = cur.queue.full = 0.;
      }
      else {
	cur.offered  = atof(f[10 + SKIP_SYSLOG_COLS]);
	cur.accepted = atof(f[12 + SKIP_SYSLOG_COLS]);
	cur.refused  = atof(f[14 + SKIP_SYSLOG_COLS]);
	cur.rejected = atof(f[16 + SKIP_SYSLOG_COLS]);
	cur.spooled  = atof(f[24 + SKIP_SYSLOG_COLS]);
	cur.missing  = atof(f[18 + SKIP_SYSLOG_COLS]);
	cur.accsize  = atof(f[20 + SKIP_SYSLOG_COLS]);
	cur.rejsize  = atof(f[22 + SKIP_SYSLOG_COLS]);
	/* parses queue values */
	/* f[34] = "198.8/200:0,0,0,0,33,67" */
	i = 0;
	pos[i++] = f[34 + SKIP_SYSLOG_COLS];
	for (p = f[34 + SKIP_SYSLOG_COLS]; *p != '\0'; p++)
	  if (*p == '/' || *p == ':' || *p == ',') {
	    *p = '\0';
	    pos[i++] = p + 1;
	  }
	cur.queue.length     = atof(pos[0]);
	cur.queue.max        = atof(pos[1]);
	cur.queue.empty      = atof(pos[2]);
	cur.queue.quartile0  = atof(pos[3]);
	cur.queue.quartile25 = atof(pos[4]);
	cur.queue.quartile50 = atof(pos[5]);
	cur.queue.quartile75 = atof(pos[6]);
	cur.queue.full       = atof(pos[7]);
      }
      /* All the connections at once */
      interpolate(&cur, peers, tm, final, id, WAY_OUT);
    }
  }
  return;
}

void parse_innd(stat_peers_t ***peers, time_t tm, char **f) {
  stat_feed_t      cur;
  char            *p;
  int              id;

  /* get the connection number of the process producing this log line */
  id = 0;
  if ((p = strchr(f[5 + SKIP_SYSLOG_COLS], ':')) != NULL) {
    *p = '\0';
    id = atoi(p + 1);
  }
  if (f[7 + SKIP_SYSLOG_COLS] == NULL) /* skip "innd: innfeed! closed" */
    return;
  if (streq(f[7 + SKIP_SYSLOG_COLS], "lost")) /* skip "innd: innfeed!xx closed lost" */
    return;
  /* process innd 'checkpoint' and 'closed' lines */
  if (streq(f[6 + SKIP_SYSLOG_COLS], "checkpoint") || streq(f[6 + SKIP_SYSLOG_COLS], "closed")) {
    int final = 0;

    if (f[6 + SKIP_SYSLOG_COLS][1] == 'l')
      final = 1;
    cur.date      = tm;
    cur.seconds   = atol(f[8 + SKIP_SYSLOG_COLS]);
    cur.offered   = 0.; /* not used */
    cur.accepted  = atof(f[10 + SKIP_SYSLOG_COLS]);
    cur.refused   = atof(f[12 + SKIP_SYSLOG_COLS]);
    cur.rejected  = atof(f[14 + SKIP_SYSLOG_COLS]);
    cur.duplicate = atof(f[16 + SKIP_SYSLOG_COLS]);
    cur.spooled   = 0.; /* not used */
    cur.missing   = 0.; /* not used */
    cur.accsize   = atof(f[19 + SKIP_SYSLOG_COLS]);
    cur.rejsize   = atof(f[22 + SKIP_SYSLOG_COLS]);

    interpolate(&cur, peers, tm, final, id, WAY_IN);
  }
}

/* parse a line */
int parse_one(char *line, stat_peers_t ***peers) {
  char      *p;
  int        app;
  time_t     tm, otm;
  struct tm *ltm;

  /* split the line into parts.
     Eg: "Dec 16 21:47:00 newsfeed innfeed[24894]: peer:1 connected"
     0-2 date ("Dec 16 21:47:00")
     3.  hostname ("newsfeed")
     4.  process ("innfeed[24894]")
     5-x data ("peer:1 connected")
  */
  split(line, &f);

  /* skip non 'innd' or 'innfeed' lines */
  if (strstr(f[4], "innfeed[") == f[4])
    app = APP_INNFEED;
  else if (strstr(f[4], "innd:") == f[4])
    app = APP_INND;
  else
    return 0;

  /* get the date in time_t format (need to compute the year) */
  f[2][2] = '\0';
  f[2][5] = '\0';
  otm = time(NULL);
  ltm = localtime(&otm);
  p = strstr(MONTHS, f[0]);
  ltm->tm_mon = (p - MONTHS) / 3;
  ltm->tm_mday = atoi(f[1]);
  ltm->tm_sec  = atoi(f[2] + 6);
  ltm->tm_min  = atoi(f[2] + 3);
  ltm->tm_hour = atoi(f[2]);
  tm = mktime(ltm);
  if (tm > otm) {
    ltm->tm_year--;
    tm = mktime(ltm);
  }
  if (min_date == 0)
    min_date = tm;
  max_date = tm;
  switch (app) {
  case APP_INNFEED:
    parse_innfeed(peers, tm, f);
    break;
  case APP_INND:
    parse_innd(peers, tm, f);
    break;
  default:
    return 0;
  }
  return 1;
}

/* parse the log file */
int parse(FIO *fio, stat_peers_t ***peers) {
  char          *p;
  stat_peers_t **q;
  int            i;

  for (;;) {
    p = fio_readline(fio);
    if (p == NULL)
      break;
    parse_one(p, peers);
  }
  for (i = 0, q = *peers; *q != NULL; q++, i++);
  return i;
}

void free_data(stat_peers_t **data) {
  stat_peers_t     **p;
  stat_feed_prev_t **q;
  stat_feed_t      **r;

  for (p = data; *p != NULL; p++) {
    for (q = (*p)->prevs; *q != NULL; q++)
      free(*q);
    for (r = (*p)->data; *r != NULL; r++)
      free(*r);
    free((*p)->data);
    free((*p)->prevs);
    free((*p)->name);
    free(*p);
  }
  free(data);
}
